/* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil -*- */
/*
 * Copyright (c) 1994 by the University of Southern California
 *
 * EXPORT OF THIS SOFTWARE from the United States of America may
 *     require a specific license from the United States Government.
 *     It is the responsibility of any person or organization contemplating
 *     export to obtain such a license before exporting.
 *
 * WITHIN THAT CONSTRAINT, permission to copy, modify, and distribute
 *     this software and its documentation in source and binary forms is
 *     hereby granted, provided that any documentation or other materials
 *     related to such distribution or use acknowledge that the software
 *     was developed by the University of Southern California.
 *
 * DISCLAIMER OF WARRANTY.  THIS SOFTWARE IS PROVIDED "AS IS".  The
 *     University of Southern California MAKES NO REPRESENTATIONS OR
 *     WARRANTIES, EXPRESS OR IMPLIED.  By way of example, but not
 *     limitation, the University of Southern California MAKES NO
 *     REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY
 *     PARTICULAR PURPOSE. The University of Southern
 *     California shall not be held liable for any liability nor for any
 *     direct, indirect, or consequential damages with respect to any
 *     claim by the user or distributor of the ksu software.
 *
 * KSU was writen by:  Ari Medvinsky, ari@isi.edu
 */

#include "ksu.h"
#include <grp.h>
#include <sys/stat.h>

static void auth_cleanup (FILE *);
static krb5_error_code conf_import_env (FILE *);


krb5_boolean fowner(fp)
    FILE *fp;
{
    struct stat sbuf;

    /*
     * For security reasons, file must be owned by
     * root.  Otherwise, don't grant access.
     */
    if (fstat(fileno(fp), &sbuf)) {
        return(FALSE);
    }

    if (sbuf.st_uid) {
        return(FALSE);
    }

    return(TRUE);
}

krb5_boolean check_group(grnam, lusernam)
     const char *grnam;
     const char *lusernam;
{
    struct group *gr;
    char *const *mem;

    if (!grnam || !*grnam)
        return FALSE;
    
    gr = getgrnam(grnam);
    if (gr == NULL)
        return FALSE;

    for (mem = gr->gr_mem; mem && *mem; ++mem) {
        if (!strcmp(lusernam, *mem))
            return TRUE;
    }
    return FALSE;
}

krb5_error_code open_conf(conf_fp_out)
     FILE ** conf_fp_out;
{
    int conf_flag =0;
    struct stat st_temp;
    FILE * conf_fp = 0;

    *conf_fp_out = NULL;
    conf_flag = stat(KLSU_CONF, &st_temp);
    if (!conf_flag) {
        if ((conf_fp = fopen(KLSU_CONF, "r")) == NULL) {
            /* cannot open, ignore */
            return -1;
        }
        /* conf file must be owned by target user or root */
        else if ( fowner(conf_fp) == FALSE) {
            /* wrong owner, ignore */
            auth_cleanup(conf_fp);
            return -1;
        }
        else {
            /* everything fine */
            *conf_fp_out = conf_fp;
            return 0;
        }
    }
    return -1;
}

/*
 * Given a Kerberos principal "principal", check if the associated local username
 * can become root according to the
 * authorization file /etc/klsu.conf and the #defines in ksu.h.  Returns TRUE
 * if authorized, FALSE if not authorized.
 *
 */

krb5_error_code krb5_authorization(context, principal, ok)
/* IN */
    krb5_context context;
    krb5_principal principal;
    /* OUT */
    krb5_boolean *ok;
{
    char lusername[LOGIN_NAME_MAX+1] = {0};
    krb5_error_code retval = 0;
    krb5_boolean temp_ok = FALSE;
    int conf_flag =0;
    FILE * conf_fp = 0;

    *ok =FALSE;

    retval = krb5_aname_to_localname(context, principal, LOGIN_NAME_MAX, lusername);
    if (retval)
        return retval;

    if (!*lusername)
        return NOT_AUTHORIZED;

    retval = krb5_kuserok(context, principal, lusername);
    if (!retval)
        return NOT_AUTHORIZED;

    if (strlen(SUPERADMINGRP) > 0) {
        temp_ok = check_group(SUPERADMINGRP, lusername);
    }

    if (!temp_ok && strlen(HOSTADMINGRP) > 0) {
        krb5_principal this_server;
        krb5_data *comp;

        retval = krb5_sname_to_principal(context, NULL, NULL, KRB5_NT_SRV_HST, &this_server);
        if (retval)
            return retval;
        comp = krb5_princ_component(context, this_server, 1);

        if (comp != NULL) {
            char *host, *dot, *groupnam;

            host = xmalloc(comp->length+1);
            if (host == NULL) {
                krb5_free_principal(context, this_server);
                return ENOMEM;
            }

            memcpy(host, comp->data, comp->length);
            host[comp->length] = '\0';

            dot = strchr(host, '.');
            if (dot != NULL)
                *dot = '\0';

            if (asprintf(&groupnam, HOSTADMINGRP, host) < 0) {
                groupnam = NULL;
                free(host);
                krb5_free_principal(context, this_server);
                return ENOMEM;
            }

            temp_ok = check_group(groupnam, lusername);
            free(groupnam);
            free(host);
        }

        krb5_free_principal(context, this_server);
    }

    if (!temp_ok) {
        /* if config file exists,
           also check it for authorized users */

        if (!(conf_flag = open_conf(&conf_fp))){
            conf_lookup(conf_fp, lusername, &temp_ok);
        }
    }

    if (temp_ok) {
        /* import env variables */
        if (!conf_flag) {
            if (conf_fp == NULL)
                /* conf file not yet open */
                conf_flag = open_conf(&conf_fp);

            else
                rewind(conf_fp);

        }

        if (!conf_flag) {
            conf_import_env(conf_fp);
        }
    }

    auth_cleanup(conf_fp);
        
    *ok =temp_ok;
    return 0;
}

/***********************************************************
conf_import_env imports the env variables from the config

***********************************************************/

krb5_error_code conf_import_env (fp)
    FILE *fp;
{

    krb5_error_code retval;
    char * line;
    char * ftok;
    char * lp;
    char * val;

    retval = get_line(fp, &line);
    if (retval)
        return retval;

    while (line){

        if (!line[0]) {
            /* ignore empty line */
        }
        else if (line[0] == '#') {
            /* ignore comment line */
        }
        else {
            lp = line;
            ftok = get_next_token (&lp);
            if (ftok && ((val = strchr(ftok, '=')))) {
                if (*lp) lp[-1] = ' ';
                *val = '\0';
                if(strncmp(ftok,"KRB",3)) /* changing KRB variables is not allowed */
                    set_env_var(ftok,val+1);
            }
        }

        free (line);
        retval = get_line(fp, &line);
        if (retval)
            return retval;
    }

    return 0;

}

/***********************************************************
conf_lookup looks for lusername in file fp. Spaces
before the lusername (in the file ) are not ignored
spaces after the user name are ignored. If there are
any tokens after the user name  FALSE is returned.

***********************************************************/

krb5_error_code conf_lookup (fp, lusername, found)
    FILE *fp;
    char *lusername;
    krb5_boolean *found;
{

    krb5_error_code retval;
    char * line;
    char * ftok;
    char * lp;
    krb5_boolean loc_found = FALSE;

    retval = get_line(fp, &line);
    if (retval)
        return retval;

    while (line){
        ftok = get_first_token (line, &lp);

        if (ftok){
            krb5_boolean temp_found = FALSE;

            if (!ftok[0]) {
                /* ignore empty line */
            }
            else if (ftok[0] == '#') {
                /* ignore comment line */
            }
            else if (strchr(ftok, '=')) {
                /* it's a variable declaration */
            }
            else if(ftok[0] == '%') {
                /* it's a group definition */
                if (ftok[1]) {
                    temp_found = check_group(ftok + 1, lusername);
                }
            }
            else if (!strcmp(lusername, ftok)) {
                temp_found = TRUE;
            }

            if (temp_found) {
                if( get_next_token (&lp) ){
                    /* nothing should follow name*/
                }
                else{
                    loc_found = TRUE;
                    free (line);
                    break;
                }
            }
        }

        free (line);

        retval = get_line(fp, &line);
        if (retval)
            return retval;
    }

    *found = loc_found;
    return 0;

}



/*********************************************************
   get_line - returns a line of any length.  out_line
              is set to null if eof.
*********************************************************/

krb5_error_code get_line (fp, out_line)
/* IN */
    FILE *fp;
    /* OUT */
    char **out_line;
{
    char * line, *r, *newline , *line_ptr;
    int chunk_count = 1;

    line = (char *) xcalloc (BUFSIZ, sizeof (char ));
    line_ptr = line;
    line[0] = '\0';

    while (( r = fgets(line_ptr, BUFSIZ , fp)) != NULL){
        newline = strchr(line_ptr, '\n');
        if (newline) {
            *newline = '\0';
            break;
        }
        else {
            chunk_count ++;
            if(!( line = (char *) realloc( line,
                                           chunk_count * sizeof(char) * BUFSIZ))){
                return  ENOMEM;
            }

            line_ptr = line + (BUFSIZ -1) *( chunk_count -1) ;
        }
    }

    if ((r == NULL) && (strlen(line) == 0)) {
        *out_line = NULL;
    }
    else{
        *out_line = line;
    }

    return 0;
}

/*******************************************************
get_first_token -
Expects a '\0' terminated input line .
If there are any spaces before the first token, they
will be returned as part of the first token.

Note: this routine reuses the space pointed to by line
******************************************************/

char *  get_first_token (line, lnext)
    char *line;
    char **lnext;
{

    char * lptr, * out_ptr;


    out_ptr = line;
    lptr = line;

    while (( *lptr == ' ') || (*lptr == '\t')) lptr ++;

    if (strlen(lptr) == 0) return NULL;

    while (( *lptr != ' ') && (*lptr != '\t') && (*lptr != '\0')) lptr ++;

    if (*lptr == '\0'){
        *lnext = lptr;
    } else{
        *lptr = '\0';
        *lnext = lptr + 1;
    }

    return out_ptr;
}
/**********************************************************
get_next_token -
returns the next token pointed to by *lnext.
returns NULL if there is no more tokens.
Note: that this function modifies the stream
      pointed to by *lnext and does not allocate
      space for the returned tocken. It also advances
      lnext to the next tocken.
**********************************************************/

char *  get_next_token (lnext)
    char **lnext;
{
    char * lptr, * out_ptr;


    lptr = *lnext;

    while (( *lptr == ' ') || (*lptr == '\t')) lptr ++;

    if (strlen(lptr) == 0) return NULL;

    out_ptr = lptr;

    while (( *lptr != ' ') && (*lptr != '\t') && (*lptr != '\0')) lptr ++;

    if (*lptr == '\0'){
        *lnext = lptr;
    } else{
        *lptr = '\0';
        *lnext = lptr + 1;
    }

    return out_ptr;
}

static void auth_cleanup(conf_fp)
    FILE *conf_fp;
{

    if (conf_fp)
        fclose(conf_fp);
}

